#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os, sys
import subprocess

from IBK import *

VARIANTS_SUBDIR = "variants"
DELPHIN_EXECUTABLE = "DelphinSolver"
# get absolute file path to DELPHIN solver (assumed to be inside this directory)
DELPHIN_EXECUTABLE = os.path.join(os.getcwd(), DELPHIN_EXECUTABLE)
if not os.path.exists(DELPHIN_EXECUTABLE):
	print("Missing DELPHIN Solver executable '{}'".format(DELPHIN_EXECUTABLE))
	exit(1)


def read_file(fname):
	try:
		fobj = open(fname, 'r')
		content = fobj.read()
		fobj.close()
		del fobj # release file handle
		return content
	except IOError as e:
		print(str(e))
		raise RuntimeError("Error reading/opening file '{}'".format(fname))

def write_file(fname, contents):
	try:
		fobj = open(fname, 'w')
		fobj.write(contents)
		fobj.close()
		del fobj # release file handle
	except IOError as e:
		print(str(e))
		raise RuntimeError("Error opening file '{}' for writing".format(fname))


# *** main ***

# 1. read template project file
TEMPLATE = read_file('template.d6p')

# 2. create directory for project variants
if not os.path.exists(VARIANTS_SUBDIR):
	os.mkdir(VARIANTS_SUBDIR)

# 3. loop over range of moisture contents - 0.01...0.21 and create project variations
jobs = []
for i in range(20):
	moisture_content = 0.01 + i*0.01

	# replace placeholder text
	variant = TEMPLATE.replace('${INITIAL_MOISTURE_CONTENT}','{}'.format(moisture_content))
	
	# create variant file name, file name pattern = var_xx.d6p  where xx is a zero-leading number
	d6pfile = VARIANTS_SUBDIR + "/var_{:02d}.d6p".format(i)
	
	# write the file, unless it exists already
	if not os.path.exists(d6pfile):
		write_file(d6pfile, variant)
		
	# add to list of jobs, unless project was already simulated, use command line arguments -x for windows, 
	# and -p=2 for 2 CPUs per job
	if not os.path.exists(d6pfile[:-4] + "/var/restart.bin"):
		jobs.append([DELPHIN_EXECUTABLE, '-x', '--verbosity-level=0', '-p=2', d6pfile])


# 4. run all the jobs, one after another
#for job in jobs:
#	try:
#		print(job[-1])
#		solverProcess = subprocess.Popen(job)
#		solverProcess.wait()
#	except OSError as e:
#		print(str(e))

# 4. run all the jobs in parallel
jobRunner = JobRunner(4) # 4 jobs at the same time
jobRunner.run(jobs)


# 5. read all result files and collect results

# create manager for geometry files - actually not used here, but needed by Delphin6OutputFile()
geoManager = Delphin6GeoFileManager() 

resultData = []
for i in range(20):
	moisture_content = 0.01 + i*0.01
	d6pfile = VARIANTS_SUBDIR + "/var_{:02d}.d6p".format(i)
	
	# NOTE: instead of recreating filename and variant here, one could also store this information
	#       in a list and use it here
	
	# path to result d6o file
	resultFile = d6pfile[:-4] + "/results/Moisture_content_integral.d6o"
	
	# read file
	d6o = Delphin6OutputFile()
	if not d6o.read(resultFile, geoManager):
		print("Error reading result file '{}'".format(resultFile))
		continue
	
	# get value for time index #2
	value = d6o.valuesAtIndex(2)
	
	# store in vector
	resultData.append( (moisture_content, value[0]) )


# 6. create a tsv-file with initial moisture content vs. final content

tsv = "Moisture content [m3/m3]\tFinal moisture mass [kg/m]\n"
for val in resultData:
	tsv = tsv + "{}\t{}\n".format(val[0], val[1])

write_file("results.tsv", tsv)
print("Wrote 'results.tsv'.")

	
	
